# import packages
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from torch import optim

from configilm import ConfigILM
from configilm.ConfigILM import ILMConfiguration, ILMType
from configilm.extra.BEN_DataModule_LMDB_Encoder import BENDataModule
from configilm.extra.BEN_lmdb_utils import resolve_ben_data_dir
import typer
import os
import wandb
from pytorch_lightning.loggers.wandb import WandbLogger
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.callbacks import EarlyStopping
from pytorch_lightning.callbacks import LearningRateMonitor
from torchmetrics.classification import MultilabelAveragePrecision
from torchmetrics.classification import MultilabelF1Score
from LinWarCosAnLR import LinearWarmupCosineAnnealingLR


__author__ = "Leonard Hackel - BIFOLD/RSiM TU Berlin"
os.environ["WANDB_START_METHOD"] = "thread"
wandb_api_key = os.environ["WANDB_API_KEY"]


class LitVisionEncoder(pl.LightningModule):
    """
    Wrapper around a pytorch module, allowing this module to be used in automatic
    training with pytorch lightning.
    Among other things, the wrapper allows us to do automatic training and removes the
    need to manage data on different devices (e.g. GPU and CPU).
    """

    def __init__(
        self,
        config: ConfigILM.ILMConfiguration,
        lr: float = 1e-3,
    ):
        super().__init__()
        self.lr = lr
        self.config = config
        self.model = ConfigILM.ConfigILM(config)

    def training_step(self, batch, batch_idx):
        x, y = batch
        x_hat = self.model(x)
        loss = F.binary_cross_entropy_with_logits(x_hat, y)
        self.log("train/loss", loss)
        return {"loss": loss}

    def configure_optimizers(self):
        optimizer = optim.AdamW(self.parameters(), lr=self.lr, weight_decay=0.01)

        # these are steps if interval is set to step
        max_intervals = int(self.trainer.max_epochs *
                            len(self.trainer.datamodule.train_ds) /
                            self.trainer.datamodule.batch_size)
        warmup = 10000 if max_intervals > 10000 else 100 if max_intervals > 100 else 0

        print(f"Optimizing for {max_intervals} steps with warmup for {warmup} steps")

        lr_scheduler = {
            'scheduler': LinearWarmupCosineAnnealingLR(
                optimizer,
                warmup_epochs=warmup,
                max_epochs=max_intervals,
                warmup_start_lr=self.lr / 10,
                eta_min=self.lr / 10
            ),
            'name': 'learning_rate',
            'interval': "step",
            'frequency': 1
        }
        return [optimizer], [lr_scheduler]

    def validation_step(self, batch, batch_idx):
        x, y = batch
        x_hat = self.model(x)
        loss = F.binary_cross_entropy_with_logits(x_hat, y)
        return {"loss": loss, "outputs": x_hat, "labels": y}

    def validation_epoch_end(self, outputs):
        metrics = self.get_metrics(outputs)

        self.log("val/loss", metrics["avg_loss"])
        self.log("val/f1", metrics["avg_f1_score"])
        self.log("val/mAP (Micro)", metrics["map_score"]["micro"])
        self.log("val/mAP (Macro)", metrics["map_score"]["macro"])

    def test_step(self, batch, batch_idx):
        x, y = batch
        x_hat = self.model(x)
        loss = F.binary_cross_entropy_with_logits(x_hat, y)
        return {"loss": loss, "outputs": x_hat, "labels": y}

    def test_epoch_end(self, outputs):
        metrics = self.get_metrics(outputs)

        self.log("test/loss", metrics["avg_loss"])
        self.log("test/f1", metrics["avg_f1_score"])
        self.log("test/mAP (Micro)", metrics["map_score"]["micro"])
        self.log("test/mAP (Macro)", metrics["map_score"]["macro"])

    def forward(self, batch):
        # because we are a wrapper, we call the inner function manually
        return self.model(batch)

    def get_metrics(self, outputs):
        avg_loss = torch.stack([x["loss"] for x in outputs]).mean()
        logits = torch.cat([x["outputs"].cpu() for x in outputs], 0)
        labels = torch.cat(
            [x["labels"].cpu() for x in outputs], 0
        )  # Tensor of size (#samples x classes)

        f1_score = MultilabelF1Score(num_labels=self.config.classes, average=None).to(
            logits.device
        )(logits, labels)

        # calculate AP
        ap_micro = MultilabelAveragePrecision(
            num_labels=self.config.classes, average="micro"
        ).to(logits.device)(logits, labels.int())

        ap_macro = MultilabelAveragePrecision(
            num_labels=self.config.classes, average="macro"
        ).to(logits.device)(logits, labels.int())

        ap_score = {"micro": float(ap_micro), "macro": float(ap_macro)}

        avg_f1_score = float(
            torch.sum(f1_score) / self.config.classes
        )  # macro average f1 score

        return {
            "avg_loss": avg_loss,
            "avg_f1_score": avg_f1_score,
            "map_score": ap_score,
        }


def main(
        model_name: str = "mobilevit_s",
        lr: float = 1e-3,
        epochs: int = 100,
        batch_size: int = 32,
        seed: int = 42,
        data_dir: str = None,
        test_run: bool = False,
        num_workers_dataloader: int = 4
):
    if test_run:
        max_img_index = 10 * batch_size
        epochs = 10
    else:
        max_img_index = -1

    pl.seed_everything(seed, workers=True)

    img_size = 120
    channels = 10

    model_config = ILMConfiguration(
        timm_model_name=model_name,
        hf_model_name=None,
        classes=19,
        image_size=img_size,
        channels=channels,
        network_type=ILMType.VISION_CLASSIFICATION
    )

    # Key is available by wandb, project name can be chosen at will
    wandb.login(key=wandb_api_key)

    tags = ["Pretraining", model_name]
    if test_run:
        tags += ["Test Run"]
    wandb_logger = WandbLogger(project=f"LiT4RSVQA",
                               log_model=True,
                               tags=tags,  # keyword arg directly to wandb.init()
                               )

    monitor = "val/mAP (Micro)"
    monitor_str = "mAP_micro"
    # checkpointing
    checkpoint_callback = ModelCheckpoint(
        monitor="val/f1",
        dirpath="./checkpoints",
        filename=f"{wandb_logger.experiment.name}-{model_name}-seed=" + str(seed) +
                 "-epoch={epoch:03d}-" + f"{monitor_str}" + "={" + f"{monitor}" +
                 ":.3f}",
        auto_insert_metric_name=False,
        save_top_k=1,
        mode="max",
        save_last=True
    )
    early_stopping_callback = EarlyStopping(monitor=monitor, min_delta=0.00,
                                            patience=25, verbose=False, mode="max")
    lr_monitor = LearningRateMonitor(logging_interval='step')

    trainer = pl.Trainer(
        max_epochs=epochs,
        accelerator="auto",
        log_every_n_steps=5,
        logger=wandb_logger,
        check_val_every_n_epoch=5,
        callbacks=[checkpoint_callback, early_stopping_callback, lr_monitor],

    )

    model = LitVisionEncoder(config=model_config, lr=lr)
    dm = BENDataModule(
        data_dir=resolve_ben_data_dir(data_dir=data_dir),
        img_size=(channels, img_size, img_size),
        num_workers_dataloader=num_workers_dataloader,
        batch_size=batch_size,
        max_img_idx=max_img_index,
    )

    trainer.fit(model=model, datamodule=dm)
    trainer.test(model=model, datamodule=dm, ckpt_path="best")

    wandb.finish()
    print("=== Training finished ===")


if __name__ == "__main__":
    typer.run(main)
